# coding:utf-8
import numpy as np
import pandas as pd
import codecs
from scipy import misc
import copy
import random

### Constants
FILE_ENCODE = "utf-8"

### Function ###
##workerのid関連
def count_about_reviewer(comp_data, comp_name_col, comp_vote_col):
    group = comp_data.groupby(comp_name_col)
    #make creators' name encoder/decoder
    compEncoder = {}
    compId = 0
    for key,data in group:
        compEncoder[key] = compId
        compId +=1
    compDecoder = {v:k for k,v in compEncoder.items()}
    return compEncoder, compDecoder

def count_about_creator(crt_data, crt_name_col, crt_make_col):
    group = crt_data.groupby(crt_name_col)
    #make creators' name encoder/decoder
    crtEncoder = {}
    crtId = 0
    for key,data in group:
        crtEncoder[key] = crtId
        crtId +=1
    crtDecoder = {v:k for k,v in crtEncoder.items()}
    return crtEncoder, crtDecoder

##leagueの総数、leagueごとのartifact数,比較pair数
def count_about_league(crt_data, crt_make_col):
    #count artifact's num par task
    artNum = np.array([])
    taskId = 0
    while 1:
        leagueDF = crt_data[crt_data[crt_make_col].str.startswith(str(taskId)+"-")]
        if leagueDF.empty:
            break
        artNum = np.r_[artNum,len(leagueDF)]
        taskId +=1
    taskNum = taskId
    return taskNum, artNum

def count_pair_num(artifact_num_arr, task_num):
    #count pair's num par task
    pairNum = np.array([])
    i = 0
    while i < task_num:
        pairNum = np.r_[pairNum, int(misc.comb(artifact_num_arr[i],2))]
        i +=1
    return pairNum

##vote-tableを作る
def make_VoteTabelCol(task_num, art_num_arr, pair_num_arr):
    label = []
    for tIndex in range(int(task_num)):
        pair_type_id = []
        pIndex =  0
        i1 = 0
        while i1 < art_num_arr[tIndex]-1:
            i2 = i1 + 1
            while i2 < art_num_arr[tIndex]:
                pair_type_id.append("-".join([str(pIndex),str(i1),str(i2)]))
                pair_type_id.append("-".join([str(pIndex),str(i2),str(i1)]))
                pIndex += 1
                i2 += 1
            i1 += 1
        if  pIndex != int(pair_num_arr[tIndex]):
            print("Worning: VoteTableColumns")
        label += ["-".join([str(tIndex),pair_type_id[i]]) for i in range(int(pair_num_arr[tIndex])*2)]
    return label

def make_VoteTable(comp_data, comp_name_col, comp_vote_col, task_num, art_num_arr, pair_num_arr):# ful comparison dataframe
    group = comp_data.groupby(comp_name_col)[comp_vote_col]
    tableCol = make_VoteTabelCol(task_num, art_num_arr, pair_num_arr)
    rtn = pd.DataFrame(columns=tableCol)
    for key, data in group:
        arBuf = data.real
        srsBuf = pd.Series([1 for i in range(len(arBuf))], index=arBuf, name=key)
        rtn = rtn.append(srsBuf)
    rtn = rtn == 1
    return rtn

##make-tableを作る
def make_MakeTableCol(task_num, artNumArr):
    label = []
    for tIndex in range(task_num):
        artNum = artNumArr[tIndex]
        label += ["-".join([str(tIndex),str(i)]) for i in range(int(artNum))]
    return label

def make_MakeTable(crt_data, crt_name_col, crt_make_col, task_num, artNumArr):
    group = crt_data.groupby(crt_name_col)[crt_make_col]
    tableCol = make_MakeTableCol(task_num, artNumArr)
    rtn = pd.DataFrame(columns=tableCol)
    for key, data in group:
        arBuf = data.real
        srsBuf = pd.Series([1 for i in range(len(arBuf))], index=arBuf, name=key)
        rtn = rtn.append(srsBuf)
    rtn = rtn == 1
    return rtn

### Class ###
class Dataset(object):
    """コンストラクタ0"""
    def __init__(self):
        self.ComparisonDataFile = ""
        self.ReviewerNameCol = ""
        self.ReviewerVoteCol = ""
        self.CreatorDataFile = ""
        self.CreatorNameCol = ""
        self.CreatorMakeCol = ""
        #DataFrame関連
        self.ComparisonDF = pd.DataFrame()
        self.CreatorDF = pd.DataFrame()
        #workerのid関連
        self.ReviewerEncoder = {}
        self.ReviewerDecoder = {}
        self.CreatorEncoder = {}
        self.CreatorDecoder = {}
        #workerの総数
        self.ReviewerNum = 0
        self.CreatorNum = 0
        #leagueの総数、leagueごとのartifact数,比較pair数
        self.taskNum = 0
        self.artNumArr = []
        self.pairNumArr = []
        #vote-tableを作る
        #make-tableを作る
        self.MakeTable = pd.DataFrame

    """コンストラクタ1"""
    def __init__(self, comparison_file_name, comp_name_col, comp_vote_col, creator_file_name, crt_name_col, crt_make_col):
        #ファイル関連
        self.ComparisonDataFile = comparison_file_name
        self.ReviewerNameCol = comp_name_col
        self.ReviewerVoteCol = comp_vote_col
        self.CreatorDataFile = creator_file_name
        self.CreatorNameCol = crt_name_col
        self.CreatorMakeCol = crt_make_col
        #DataFrame関連
        self.ComparisonDF = pd.read_csv(codecs.open(self.ComparisonDataFile, 'r', FILE_ENCODE))
        self.CreatorDF = pd.read_csv(codecs.open(self.CreatorDataFile, 'r', FILE_ENCODE))
        #workerのid関連
        self.ReviewerEncoder, self.ReviewerDecoder = count_about_reviewer(self.ComparisonDF, self.ReviewerNameCol, self.ReviewerVoteCol)
        self.CreatorEncoder, self.CreatorDecoder = count_about_creator(self.CreatorDF, self.CreatorNameCol, self.CreatorMakeCol)
        #workerの総数
        self.ReviewerNum = len(self.ReviewerEncoder)
        self.CreatorNum = len(self.CreatorEncoder)
        #leagueの総数、leagueごとのartifact数,比較pair数
        self.taskNum, self.artNumArr = count_about_league(self.CreatorDF, self.CreatorMakeCol)
        self.pairNumArr = count_pair_num(self.artNumArr, self.taskNum)
        #vote-tableを作る
        self.VoteTable = make_VoteTable(self.ComparisonDF, self.ReviewerNameCol,
                self.ReviewerVoteCol, self.taskNum, self.artNumArr, self.pairNumArr)
        #make-tableを作る
        self.MakeTable = make_MakeTable(self.CreatorDF, self.CreatorNameCol,
                self.CreatorMakeCol, self.taskNum, self.artNumArr)

    def CutCompDataparLeague(self,league_num):
        df = self.ComparisonDF.ix[self.ComparisonDF[self.ReviewerVoteCol].str.startswith(str(league_num)+"-")]
        df = df.reset_index(drop=True)
        return df

    def CutCrtDataparLeague(self,league_num):
        df = self.CreatorDF.ix[self.CreatorDF[self.CreatorMakeCol].str.startswith(str(league_num)+"-")]
        df = df.reset_index(drop=True)
        return df

    def SetComparisonDF(self,source_df,comp_name_col,comp_vote_col):
        self.ComparisonDF = source_df
        self.ReviewerNameCol = comp_name_col
        self.ReviewerVoteCol = comp_vote_col
        self.ReviewerEncoder, self.ReviewerDecoder = count_about_reviewer(self.ComparisonDF, self.ReviewerNameCol, self.ReviewerVoteCol)
        self.ReviewerNum = len(self.ReviewerEncoder)
        self.VoteTable = make_VoteTable(self.ComparisonDF, self.ReviewerNameCol,
                self.ReviewerVoteCol, self.taskNum, self.artNumArr, self.pairNumArr)

    def SetCreatorDF(self,source_df,crt_name_col,crt_make_col):
        self.CreatorDF = source_df
        self.CreatorNameCol = crt_name_col
        self.CreatorMakeCol = crt_make_col
        self.CreatorEncoder, self.CreatorDecoder = count_about_creator(self.CreatorDF, self.CreatorNameCol, self.CreatorMakeCol)
        self.CreatorNum = len(self.CreatorEncoder)
        self.taskNum, self.artNumArr = count_about_league(self.CreatorDF, self.CreatorMakeCol)
        self.pairNumArr = count_pair_num(self.artNumArr, self.taskNum)
        self.MakeTable = make_MakeTable(self.CreatorDF, self.CreatorNameCol,
                self.CreatorMakeCol, self.taskNum, self.artNumArr)

    def RandomCutCompDataPairLevel(self, random_worker_num):
        new = copy.deepcopy(self)
        source_df = new.ComparisonDF
        df = pd.DataFrame()
        tIndex=0
        while tIndex < new.taskNum:# task-level
            pairIndex=0
            while pairIndex < new.pairNumArr[tIndex]:# pairwise-level
                bufDF = source_df.ix[source_df[new.ReviewerVoteCol].str.startswith(str(tIndex)+"-"+ str(pairIndex) +"-")]
                sampler = np.random.permutation(len(bufDF))
                bufDF = bufDF.take(sampler[:random_worker_num])# ランダムな要素に圧縮
                df = pd.concat([df,bufDF],axis=0)
                pairIndex += 1
            tIndex += 1
        df = df.reset_index(drop=True)
        new.SetComparisonDF(df,new.ReviewerNameCol,new.ReviewerVoteCol)
        return new

    def RandomCutCompDataWorkerLevel(self, random_worker_num):
        new = copy.deepcopy(self)
        source_df = new.ComparisonDF
        worker_level_group = source_df.groupby(new.ReviewerNameCol)
        regular_workers_undecoreter = worker_level_group.size()[worker_level_group.size() >= 5].index.map(lambda x:x)
        random.shuffle(regular_workers_undecoreter)#5回以上投票したワーカーをシャッフルしてワーカーの並びを作る
        regular_workers_decorater = {value:i for i,value in enumerate(regular_workers_undecoreter)}
        #5回以上投票したワーカーのみのDataFrame
        regularWorkerDF = source_df[source_df[new.ReviewerNameCol].map(lambda x: x in regular_workers_undecoreter)]
        df = pd.DataFrame([])
        tIndex = 0
        while tIndex < new.taskNum:# kiji-level
            pairIndex=0
            while pairIndex < new.pairNumArr[tIndex]:# pairwise-level
                bufDF = regularWorkerDF[regularWorkerDF[new.ReviewerVoteCol].str.startswith(str(tIndex)+"-"+ str(pairIndex) +"-")]
                bufDF = bufDF.reset_index(drop = True)
                decorateBuf =[regular_workers_decorater[name] for name in bufDF[new.ReviewerNameCol].real]
                bufSrs = pd.Series(decorateBuf, name="decorate")
                bufDF = pd.concat([bufDF,bufSrs], axis=1)
                bufDF = bufDF.sort_values(by="decorate")
                bufDF = bufDF.iloc[:random_worker_num,:2]#3がno. of reviewer par pair
                df = pd.concat([df,bufDF],axis=0)
                pairIndex += 1
            tIndex += 1
        df = df.reset_index(drop = True)
        new.SetComparisonDF(df,new.ReviewerNameCol,new.ReviewerVoteCol)
        return new

    def SearchCreatorName(self, task_id, artifact_id):
        s = "-".join([str(task_id),str(artifact_id)])
        return self.MakeTable.ix[self.MakeTable[s],s].index[0]

    def SearchCreatorId(self, task_id, artifact_id):
        return self.CreatorEncoder[self.SearchCreatorName(task_id, artifact_id)]
