# coding:utf-8
from inputData import Dataset
import numpy as np
import pandas as pd
import time

### Constants
threshold = 0.00001
max_i = 500

### 実行function
def calcPH(input_dataset, output_file='None', lam=0., iter_file='None'):
    if isinstance(input_dataset, Dataset):
        dataset = input_dataset
    else:
        print('type error: calcPH input type error. expected original class:Dataset')
    # リーグごとにpairwiseHITSスコアを出す
    A, R, C = getpairwiseHITSscore(dataset,lam=lam, iter_file=iter_file)
    # ランキングリストを出す
    rankCol = ["taskNum","creator","pairwiseHITSscore"]
    rdf = pd.DataFrame(columns=rankCol) #resultDataFrame
    tIndex=0
    while tIndex < dataset.taskNum:# kiji-level
        result = A[tIndex]
        artNum = dataset.artNumArr[tIndex]
        dfBuf = pd.DataFrame({
            rankCol[0]:[tIndex for x in range(int(artNum))],
            rankCol[1]:[dataset.SearchCreatorName(tIndex,artIndex) for artIndex in range(int(artNum))],
            rankCol[2]:result
            })
        rdf = pd.concat([rdf,dfBuf],axis=0)
        tIndex += 1
    rdf = rdf.reset_index(drop = True)

    if output_file != 'None':
        rdf.to_csv(output_file,index=False)
        print("OK: {0} is maked.".format(output_file))
    return rdf

### main function
def getpairwiseHITSscore(data_set,lam=0., iter_file='None'):# 比較全体のDataFrame,最適化結果を返す
    #initialize
    A = initialize_A(data_set.taskNum, data_set.artNumArr)#list
    R = np.ones(data_set.ReviewerNum)#reviewerの能力は一律で1とした, R[i]はrVoteDF.index[i]に相当
    C = np.zeros(data_set.CreatorNum)#creatorの能力は一律で0とした, R[i]はcMakeDF.index[i]に相当
    ListP, ListInvP = make_P(data_set.taskNum, data_set.artNumArr, data_set.pairNumArr)#list

    #optimize-epoch
    current = np.r_[np.concatenate(A), R, C]
    start = time.time()
    for i in range(1,max_i+1):
        previous = current
        A,R,C = cycle(A, R, C, ListP, ListInvP, lam, data_set)
        current = np.r_[np.concatenate(A), R, C]
        diff = np.linalg.norm(previous - current)
        if i == max_i:
            print("cycle: {0}, diff={1:.5f}".format(i,diff))
        if diff <= threshold:
            if iter_file != 'None':
                f = open(iter_file, "a")
                f.write("{0}\n".format(i))
            break
    elapsed_time = time.time() - start
    print("elapsed_time:{0}[sec]".format(elapsed_time))
    return A,R,C

### Function
## 最適化1サイクル
def cycle(A, R, C, ListP, ListInvP, lam, data_set):
    #Aの更新と各ペアの誤差計算
    a, diff_a = updata_A(A, R, C, ListP, ListInvP, lam, data_set)
    #Rの更新
    r = updataR(R, diff_a, data_set)
    #Cの更新
    c = updataC(C, a, data_set)
    #R,Cの正則化
    r = normalize(r)
    c = normalize(c)
    """
    r,c = doubleNormalize(r,c)
    """
    return a,r,c

##前処理と初期化
def initialize_A(task_num, art_num_arr):
    A = []
    for tIndex in range(int(task_num)):
        a = np.zeros(int(art_num_arr[tIndex]))
        A.append(a)
    return A

def make_P(task_num, art_num_arr, pair_num_arr):
    listP = []
    listInvP = []
    for tIndex in range(int(task_num)):
        P = np.zeros([int(pair_num_arr[tIndex]),int(art_num_arr[tIndex])])
        i1 = 0
        pIndex = 0
        while i1 < art_num_arr[tIndex]-1:
            i2 = i1 + 1
            while i2 < art_num_arr[tIndex]:
                P[pIndex,i1] = 1
                P[pIndex,i2] = -1
                pIndex += 1
                i2 += 1
            i1 += 1
        if  pIndex != int(pair_num_arr[tIndex]): print("Worning: make_P")
        listP.append(P)
        listInvP.append(np.linalg.pinv(P))
    return listP,listInvP

##各更新
def updata_A(A, R, C, ListP, ListInvP, lam, data_set):
    a = A
    Diff_A = np.array([])
    tIndex = 0
    while tIndex < data_set.taskNum:
        pairNum = data_set.pairNumArr[tIndex]
        a_i = a[tIndex]#task_iのartifact
        P = ListP[tIndex]#task_iのP
        invP = ListInvP[tIndex]#task_iのinvP
        #sum(r)とsum(r')の計算
        supe = superiorSum(R,data_set,tIndex)
        infe = inferiorSum(R,data_set,tIndex)
        #task_iについての連立方程式を解く。PA=sum(r)-sum(r')
        a_i = invP.dot(supe-infe)
        #A_iの更新。A=(1-λ)A*+λC
        a_i = (1-lam)*a_i + lam*creator_i(C,data_set, tIndex)
        #task_iについてのairtifactのquality差を計算、a-a'の値
        diff_a = calcDiff_A(a_i,P, pairNum)
        Diff_A = np.r_[Diff_A,diff_a]
        a[tIndex] = a_i#task_iについて更新
        tIndex +=1
    return a,Diff_A


"""#v3.0
def updataR(R, diff_a, data_set):
    r = R
    vote_df = data_set.VoteTable
    Index = 0
    while Index < data_set.ReviewerNum:
        mask = vote_df.ix[vote_df.index[Index]].real
        r[Index] = diff_a[mask].sum()/mask.sum()
        Index +=1
    return r
"""
#v3.1
def updataR(R, diff_a, data_set):
    r = R
    vote_df = data_set.VoteTable
    Index = 0
    while Index < data_set.ReviewerNum:
        mask = vote_df.ix[vote_df.index[Index]].real
        r_local = diff_a[mask]
        r[Index] = float((r_local > 0).sum())/mask.sum()
        Index +=1
    return r

def updataC(C, a, data_set):
    c = C
    create_df = data_set.MakeTable
    Index = 0
    while Index < data_set.CreatorNum:
        mask = create_df.ix[create_df.index[Index]].real
        created = create_df.ix[create_df.index[Index]].sum()
        c[Index] = np.concatenate(a)[mask].sum()/created
        Index += 1
    return c

## 正則化
def normalize(V):
    v = V
    buf = v**2
    v = v/np.sqrt(buf.sum())
    return v

## ダブル正則化
def doubleNormalize(R,C):
    mix = np.r_[R,C]
    buf = mix**2
    r = R/np.sqrt(buf.sum())
    c = C/np.sqrt(buf.sum())
    return r,c

## Aの更新、Diff_Aの計算に関するfunction
def voteSumation(type_of_vote, reviewer_reliability_arr, vote_df):
        mask = vote_df[type_of_vote].real
        return reviewer_reliability_arr[mask].sum()

def superiorSum(R,data_set,task_id):
    voteType = data_set.VoteTable.columns
    votes = voteType[voteType.str.startswith(str(task_id)+"-")]
    pairNum = data_set.pairNumArr[task_id]
    sup = np.array([])
    for num in range(int(pairNum)):
        n = 2*num
        supVote = votes[n]
        sup = np.r_[sup,voteSumation(supVote, R, data_set.VoteTable)]
    return sup

def inferiorSum(R,data_set,task_id):
    voteType = data_set.VoteTable.columns
    votes = voteType[voteType.str.startswith(str(task_id)+"-")]
    pairNum = data_set.pairNumArr[task_id]
    inf = np.array([])
    for num in range(int(pairNum)):
        n = 2*num + 1
        infVote = votes[n]
        inf = np.r_[inf,voteSumation(infVote, R, data_set.VoteTable)]
    return inf

def creator_i(C,data_set, task_id):
    create_df = data_set.MakeTable
    local_c = np.array([])
    for art_id in range(int(data_set.artNumArr[task_id])):
        mask = create_df[str(task_id)+"-"+str(art_id)].real
        local_c = np.r_[local_c, C[mask]]
    return local_c

def calcDiff_A(A_i,P, pair_num):
    su = P.dot(A_i)
    diff = np.c_[su,-su]
    return diff.reshape(int(pair_num)*2)
